use serde::{Deserialize, Serialize};
use surrealdb::{
    engine::remote::ws::{Client, Ws},
    opt::auth::Root,
    sql::Thing,
    Surreal,
};
use tracing::{debug, error, info, instrument};
use url::Url;

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Website {
    /// The url that this data is found at
    site: Url,
    /// The url as defined in the <a> tag
    href: Url,
    /// Wether or not this link has been crawled yet
    crawled: bool,
    /// Wether or not the href was doctored
    doctored_href: bool,
    original_href: Option<String>,
}

impl Website {
    /// Creates a blank site (assumes that url param is site's root)
    pub fn new(url: &str, href: &str, crawled: bool) -> Self {
        let mut new = Self::from(url);
        new.crawled = crawled;
        new.original_href = Some(href.to_string());
        new.href =
            match Url::parse(href) {
                Ok(e) => e,
                Err(e) => {
                    match e {
                        url::ParseError::RelativeUrlWithoutBase => {
                            // Try to combine the scheme_host and href to get a useable domain
                            new.doctored_href = true;

                            let url = if !url.ends_with('/') && !href.starts_with('/') {
                                format!("{url}/{href}")
                            } else {
                                format!("{url}{href}")
                            };

                            // paste the domain onto the begining of the href
                            Url::parse(&url).map_or_else(|err| {
                                debug!("Parsing {url} with {href}");
                                error!("{err} Failed to parse href into url on second try. Aborting");
                                panic!("See error logs for more info.");
                            }, |ok| ok)
                        }
                        _ => {
                            error!("{e}");
                            panic!("See error logs for more info.");
                        }
                    }
                }
            };
        new
    }
    pub fn crawled(&mut self) {
        self.crawled = true
    }
    pub fn href_str(&self) -> &str {
        self.href.as_str()
    }
    pub fn site(&self) -> String {
        self.site.to_string()
    }
    pub fn domain_str(&self) -> &str {
        self.site.as_str()
    }
    #[instrument(skip_all)]
    pub async fn store(&mut self, db: &Surreal<Client>) {
        // is root record?
        if self.href.path() == "/" {
            // Upsert is create or update
            // Whereas Update is just update
            let record = ("website", &self.href.to_string());

            let crawled = if let Some(old) = db.select(record).await.unwrap() {
                let old: Website = old; // infer type
                old.crawled
            } else {false};

            if !self.crawled {self.crawled = crawled};

            match db.upsert(record).content(self.clone()).await {
                Ok(e) => {
                    if let Some(a) = &e {
                        let _: &Record = a;
                    }
                }
                Err(e) => {
                    error!("{}", e);
                },
            };
        } else {
            let _: Option<Record> = match db.create("website").content(self.clone()).await {
                Ok(e) => {
                    if let Some(a) = &e {
                        let _: &Record = a;
                    }
                    e
                }
                Err(_) => todo!(),
            };
        }
    }
}

impl From<&str> for Website {
    /// site == href, crawled = false
    fn from(value: &str) -> Self {
        let site = match Url::parse(value) {
            Ok(a) => a,
            Err(_) => todo!(),
        };
        Self {
            href: site.clone(),
            crawled: false,
            site,
            doctored_href: false,
            original_href: None,
        }
    }
}

#[derive(Debug, Serialize)]
pub struct Email {
    pub email: String,
}

#[derive(Debug, Deserialize)]
pub struct Record {
    #[allow(dead_code)]
    pub id: Thing,
}

pub async fn connect() -> surrealdb::Result<Surreal<Client>> {
    // Connect to the server
    let db = Surreal::new::<Ws>("127.0.0.1:8000").await?;

    // Signin as a namespace, database, or root user
    db.signin(Root {
        username: "root",
        password: "root",
    })
    .await?;

    // Select a specific namespace / database
    db.use_ns("test").use_db("test").await?;

    Ok(db)
}
